/*
 * Copyright (C) 2006-2010 Michael Niedermayer <michaelni@gmx.at>
 *               2010      James Darnley <james.darnley@gmail.com>
 *
 * This file is part of Libav.
 *
 * Libav is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Libav; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <libintl.h>
#define _(str) gettext(str)		///< gettext shortcut
#define _N(str) str			///< gettext_noop shortcut
#include "vaapi_yadif.h"
#include "yadif.h"
#include "misc.h"

#define CHECK(j)\
    {   int score = FFABS(cur[mrefs-1+(j)] - cur[prefs-1-(j)])\
                  + FFABS(cur[mrefs  +(j)] - cur[prefs  -(j)])\
                  + FFABS(cur[mrefs+1+(j)] - cur[prefs+1-(j)]);\
        if (score < spatial_score) {\
            spatial_score= score;\
            spatial_pred= (cur[mrefs  +(j)] + cur[prefs  -(j)])>>1;\

#define FILTER \
    for (x = 0;  x < w; x++) { \
        int c = cur[mrefs]; \
        int d = (prev2[0] + next2[0])>>1; \
        int e = cur[prefs]; \
        int temporal_diff0 = FFABS(prev2[0] - next2[0]); \
        int temporal_diff1 =(FFABS(prev[mrefs] - c) + FFABS(prev[prefs] - e) )>>1; \
        int temporal_diff2 =(FFABS(next[mrefs] - c) + FFABS(next[prefs] - e) )>>1; \
        int diff = FFMAX3(temporal_diff0>>1, temporal_diff1, temporal_diff2); \
        int spatial_pred = (c+e)>>1; \
        int spatial_score = FFABS(cur[mrefs-1] - cur[prefs-1]) + FFABS(c-e) \
                          + FFABS(cur[mrefs+1] - cur[prefs+1]) - 1; \
 \
        CHECK(-1) CHECK(-2) }} }} \
        CHECK( 1) CHECK( 2) }} }} \
 \
        if (mode < 2) { \
            int b = (prev2[2*mrefs] + next2[2*mrefs])>>1; \
            int f = (prev2[2*prefs] + next2[2*prefs])>>1; \
            int max = FFMAX3(d-e, d-c, FFMIN(b-c, f-e)); \
            int min = FFMIN3(d-e, d-c, FFMAX(b-c, f-e)); \
 \
            diff = FFMAX3(diff, min, -max); \
        } \
 \
        if (spatial_pred > d + diff) \
           spatial_pred = d + diff; \
        else if (spatial_pred < d - diff) \
           spatial_pred = d - diff; \
 \
        dst[0] = spatial_pred; \
 \
        dst++; \
        cur++; \
        prev++; \
        next++; \
        prev2++; \
        next2++; \
    }

static void filter_line_c(uint8_t *dst,
                          uint8_t *prev, uint8_t *cur, uint8_t *next,
                          int w, int prefs, int mrefs, int parity, int mode)
{
    int x;
    uint8_t *prev2 = parity ? prev : cur ;
    uint8_t *next2 = parity ? cur  : next;

    FILTER
}

static void nv12ToUV(uint8_t *dstU, uint8_t *dstV, uint8_t *src, int width)
{
	int i;
	for(i=0;i<width;i++) {
		dstU[i]=src[i*2];
		dstV[i]=src[i*2+1];
	}
}

static void UVTonv12(uint8_t *dst, uint8_t *srcU, uint8_t *srcV, int width)
{
	int i;
	for(i=0;i<width;i++) {
		dst[i*2]=srcU[i];
		dst[i*2+1]=srcV[i];
	}
}

#define YADIF_PITCH(w,a) (w%a==0) ? w : w+a-(w%a)

static void CopyVAImg2YadifImg_alloc(VADisplay *display, VAImage *src, vaapi_yadif_img *dst)
{
	uint8_t *srcp, *srcuv, *dstU, *dstV;
	int i;

	if (vaMapBuffer(display, src->buf, (void**)&srcp) != VA_STATUS_SUCCESS) {
		Error(_("video/vaapi: can't map image buffer\n"));
    }

	dst->width[0] = src->width;
	dst->width[1] = dst->width[2] = dst->width[0]>>1;
	dst->height[0] = src->height;
	dst->height[1] = dst->height[2] = dst->height[0]>>1;
	if(src->format.fourcc == VA_FOURCC_NV12) {
		dst->pitch[0] = src->pitches[0];
		dst->pitch[2] = dst->pitch[1] = YADIF_PITCH(dst->width[1],16);
		dst->yuvsize = dst->pitch[0]*dst->height[0]+dst->pitch[1]*dst->height[1]+dst->pitch[2]*dst->height[2];
		dst->yuv[0] = malloc(dst->yuvsize);
		dst->yuv[1] = dst->yuv[0]+dst->pitch[0]*dst->height[0];
		dst->yuv[2] = dst->yuv[1]+dst->pitch[1]*dst->height[1];
		memcpy(dst->yuv[0], srcp+src->offsets[0], src->height*src->pitches[0]);
		srcuv=srcp+src->offsets[1];
		dstU=dst->yuv[1];
		dstV=dst->yuv[2];
		for(i=0;i<dst->height[1];i++) {
#if HAVE_SSSE3
			vdr_plg_shddev_nv12ToUV_ssse3(dstU,dstV,srcuv,dst->width[1]);
#elif HAVE_SSE2
			vdr_plg_shddev_nv12ToUV_sse2(dstU,dstV,srcuv,dst->width[1]);
#else
			nv12ToUV(dstU,dstV,srcuv,dst->width[1]);
#endif
			srcuv+=src->pitches[1];
			dstU+=dst->pitch[1];
			dstV+=dst->pitch[2];
		}
	}
	if (vaUnmapBuffer(display, src->buf) != VA_STATUS_SUCCESS) {
		Error(_("video/vaapi: can't unmap image buffer\n"));
    }
}

static void CopyVAImg2YadifImg(VADisplay *display, VAImage *src, vaapi_yadif_img *dst)
{
	uint8_t *srcp, *srcuv, *dstU, *dstV;
	int i;
#if YADIF_TICKS
	uint32_t tick1, tick2, tick3, tick4, tick2a, tick2b, tick2c;
	tick1 = GetMsTicks();
#endif
#if YADIF_PREFETCH
	srcp = dst->prefetch;
#else
	if (vaMapBuffer(display, src->buf, (void**)&srcp) != VA_STATUS_SUCCESS) {
		Error(_("video/vaapi: can't map image buffer\n"));
    }
#endif
#if YADIF_TICKS
	tick2 = GetMsTicks();
#endif
	if(src->format.fourcc == VA_FOURCC_NV12) {
#if YADIF_TICKS
		tick2a = GetMsTicks();
#endif
		memcpy(dst->yuv[0], srcp+src->offsets[0], src->height*src->pitches[0]);
#if YADIF_TICKS
		tick2b = GetMsTicks();
#endif
		srcuv=srcp+src->offsets[1];
		dstU=dst->yuv[1];
		dstV=dst->yuv[2];
		for(i=0;i<dst->height[1];i++) {
#if HAVE_SSSE3
			vdr_plg_shddev_nv12ToUV_ssse3(dstU,dstV,srcuv,dst->width[1]);
#elif HAVE_SSE2
			vdr_plg_shddev_nv12ToUV_sse2(dstU,dstV,srcuv,dst->width[1]);
#else
			nv12ToUV(dstU,dstV,srcuv,dst->width[1]);
#endif
			srcuv+=src->pitches[1];
			dstU+=dst->pitch[1];
			dstV+=dst->pitch[2];
		}
#if YADIF_TICKS
		tick2c = GetMsTicks();
#endif
	}
#if YADIF_TICKS
	tick3 = GetMsTicks();
#endif
	if (vaUnmapBuffer(display, src->buf) != VA_STATUS_SUCCESS) {
		Error(_("video/vaapi: can't unmap image buffer\n"));
    }
#if YADIF_TICKS
	tick4 = GetMsTicks();
//	printf("Va2ya Times: Map: %i (memcpy: %i (align: %i %i %i) nv12ToUV: %i) Copy: %i Unmap: %i - Total: %i\n", tick2-tick1, tick2b-tick2a, (uint64_t)(dst->yuv[0])%32, (uint64_t)(srcp+src->offsets[0])%32 , ((int64_t)(dst->yuv[0])-(int64_t)(srcp+src->offsets[0]))%128,tick2c-tick2b, tick3-tick2, tick4-tick3, tick4-tick1);
#endif
}

static void CopyYadifImg2VAImg(VADisplay *display, vaapi_yadif_img *src, VAImage *dst)
{
	uint8_t *dstp, *dstuv, *srcU, *srcV;
	int i;
	if(vaMapBuffer(display, dst->buf, (void**)&dstp) != VA_STATUS_SUCCESS) {
		Error(_("video/vaapi: can't map image buffer\n"));
    }
	if(dst->format.fourcc == VA_FOURCC_NV12) {
		memcpy(dstp+dst->offsets[0], src->yuv[0], src->height[0]*src->pitch[0]);
		dstuv=dstp+dst->offsets[1];
		srcU=src->yuv[1];
		srcV=src->yuv[2];
		for(i=0;i<src->height[1];i++) {
#if HAVE_SSSE3
			vdr_plg_shddev_UVTonv12_ssse3(dstuv,srcU,srcV,src->width[1]);
#else
			UVTonv12(dstuv,srcU,srcV,src->width[1]);
#endif
			dstuv+=dst->pitches[1];
			srcU+=src->pitch[1];
			srcV+=src->pitch[2];
		}
	}
	if (vaUnmapBuffer(display, dst->buf) != VA_STATUS_SUCCESS) {
		Error(_("video/vaapi: can't unmap image buffer\n"));
    }
#if YADIF_TICKS
	//printf("ya2Va Align: %i %i %i\n", (uint64_t)(src->yuv[0])%32, (uint64_t)(dstp+dst->offsets[0])%32 , ((int64_t)(src->yuv[0])-(int64_t)(dstp+dst->offsets[0]))%128);
#endif
}

static void InitVaapiYadif(VADisplay *display, VAImage *src, vaapi_yadif_img *buf_imgs[])
{
	buf_imgs[0]=malloc(5*sizeof(vaapi_yadif_img));
	//printf("init yadif1\n");
	buf_imgs[1]=((void*)buf_imgs[0])+sizeof(vaapi_yadif_img);
	buf_imgs[2]=((void*)buf_imgs[1])+sizeof(vaapi_yadif_img);
	buf_imgs[3]=((void*)buf_imgs[2])+sizeof(vaapi_yadif_img);
	buf_imgs[4]=((void*)buf_imgs[3])+sizeof(vaapi_yadif_img);
	//printf("init yadif2\n");
	CopyVAImg2YadifImg_alloc(display,src,buf_imgs[0]);
	//printf("init yadif3\n");
	memcpy(buf_imgs[1],buf_imgs[0],sizeof(vaapi_yadif_img));
	memcpy(buf_imgs[2],buf_imgs[0],sizeof(vaapi_yadif_img));
	buf_imgs[1]->yuv[0] = malloc(buf_imgs[1]->yuvsize);
	buf_imgs[2]->yuv[0] = malloc(buf_imgs[2]->yuvsize);
	memcpy(buf_imgs[1]->yuv[0],buf_imgs[0]->yuv[0],buf_imgs[1]->yuvsize);
	memcpy(buf_imgs[2]->yuv[0],buf_imgs[0]->yuv[0],buf_imgs[2]->yuvsize);
	buf_imgs[1]->yuv[1] = buf_imgs[1]->yuv[0]+buf_imgs[1]->pitch[0]*buf_imgs[1]->height[0];
	buf_imgs[1]->yuv[2] = buf_imgs[1]->yuv[1]+buf_imgs[1]->pitch[1]*buf_imgs[1]->height[1];
	buf_imgs[2]->yuv[1] = buf_imgs[2]->yuv[0]+buf_imgs[2]->pitch[0]*buf_imgs[2]->height[0];
	buf_imgs[2]->yuv[2] = buf_imgs[2]->yuv[1]+buf_imgs[2]->pitch[1]*buf_imgs[2]->height[1];
	memcpy(buf_imgs[3],buf_imgs[0],sizeof(vaapi_yadif_img));
	memcpy(buf_imgs[4],buf_imgs[0],sizeof(vaapi_yadif_img));
	buf_imgs[3]->yuv[0] = malloc(buf_imgs[3]->yuvsize);
	buf_imgs[4]->yuv[0] = malloc(buf_imgs[4]->yuvsize);
	buf_imgs[3]->yuv[1] = buf_imgs[3]->yuv[0]+buf_imgs[3]->pitch[0]*buf_imgs[3]->height[0];
	buf_imgs[3]->yuv[2] = buf_imgs[3]->yuv[1]+buf_imgs[3]->pitch[1]*buf_imgs[3]->height[1];
	buf_imgs[4]->yuv[1] = buf_imgs[4]->yuv[0]+buf_imgs[4]->pitch[0]*buf_imgs[4]->height[0];
	buf_imgs[4]->yuv[2] = buf_imgs[4]->yuv[1]+buf_imgs[4]->pitch[1]*buf_imgs[4]->height[1];
}

#if YADIF_MT
void *VaapiYadif_thread(void *targ)
{
	yadif_thread_args *arg = (yadif_thread_args*) targ;
	int i, y;
	uint8_t  **_prev, **_cur, **_next, **_outA, **_outB;
	_outA = arg->buf_imgs[3]->yuv;
	_outB = arg->buf_imgs[4]->yuv;
	int tff = arg->tff;
	int skip_spatial = arg->skip_spatial;
	while(arg->y[0]>=0)
	{
		_prev = arg->buf_imgs[0]->yuv;
		_cur = arg->buf_imgs[1]->yuv;
		_next = arg->buf_imgs[2]->yuv;

		for (i = 0; i < 3; i++) {
	        int w = arg->w[i];
	        int h = arg->h[i];
	        int refs = arg->refs[i];
			int y2 = arg->y2[i];
	        for (y = arg->y[i]; y < y2; y++) {
	            if ((y ^ tff) & 1) {
	                uint8_t *prev = &_prev[i][y*refs];
	                uint8_t *cur  = &_cur[i][y*refs];
	                uint8_t *next = &_next[i][y*refs];
#if HAVE_SSSE3
	                vdr_plg_shddev_yadif_filter_line_ssse3(&_outB[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#elif HAVE_SSE2
	                vdr_plg_shddev_yadif_filter_line_sse2(&_outB[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#else
	                filter_line_c(&_outB[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#endif
	                memcpy(&_outA[i][y*refs], cur, w);
	            } else {
	                uint8_t *prev = &_prev[i][y*refs];
	                uint8_t *cur  = &_cur[i][y*refs];
	                uint8_t *next = &_next[i][y*refs];
#if HAVE_SSSE3
	                vdr_plg_shddev_yadif_filter_line_ssse3(&_outA[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#elif HAVE_SSE2
	                vdr_plg_shddev_yadif_filter_line_sse2(&_outA[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#else
	                filter_line_c(&_outA[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#endif
	                memcpy(&_outB[i][y*refs], cur, w);
	            }
	        }
	    }
#if HAVE_MMX
	    __asm__ volatile("emms \n\t" : : : "memory");
#endif
		pthread_barrier_wait(arg->barrier);
		pthread_mutex_lock(arg->condlock);
		pthread_cond_wait(arg->cond,arg->condlock);
		pthread_mutex_unlock(arg->condlock);
	}
	pthread_exit(NULL);
}
#endif

#if YADIF_MT
void VaapiYadif (VADisplay *display, VAImage *src, VAImage *dst1,
	VAImage *dst2, vaapi_yadif_img *buf_imgs[], yadif_threads *threads, int tff, int skip_spatial)
#else
void VaapiYadif (VADisplay *display, VAImage *src, VAImage *dst1,
	VAImage *dst2, vaapi_yadif_img *buf_imgs[], int tff, int skip_spatial)
#endif
{
	vaapi_yadif_img *tmp;
	int y, i;
#if YADIF_TICKS
	uint32_t tick1, tick2, tick3, tick4, tick1a, tick1b;
	tick1 = GetMsTicks();
#endif
	skip_spatial<<=1;
	if(buf_imgs[2]==NULL) {
		InitVaapiYadif(display, src, buf_imgs);
#if YADIF_MT
		int ysum[3];
		int ystep[3];
		pthread_attr_t attr;
		ysum[2] = ysum[1] = ysum[0] = 0;
		ystep[0] = buf_imgs[3]->height[0]/YADIF_MT_THREADS;
		ystep[1] = buf_imgs[3]->height[1]/YADIF_MT_THREADS;
		ystep[2] = buf_imgs[3]->height[2]/YADIF_MT_THREADS;

		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
		pthread_cond_init(&threads->cond, NULL);
		pthread_mutex_init(&threads->condlock,NULL);
		pthread_barrier_init(&threads->barrier,NULL,YADIF_MT_THREADS+1);
		for (i=0;i<YADIF_MT_THREADS;i++) {
			threads->args[i].condlock = &threads->condlock;
			threads->args[i].cond = &threads->cond;
			threads->args[i].barrier = &threads->barrier;
			threads->args[i].buf_imgs = buf_imgs;
			threads->args[i].tff = tff;
			threads->args[i].skip_spatial = skip_spatial;
			for (y=0;y<3;y++) {
				threads->args[i].w[y] = buf_imgs[3]->width[y];
				threads->args[i].refs[y] = buf_imgs[3]->pitch[y];
				threads->args[i].h[y] = buf_imgs[3]->height[y];
				threads->args[i].y[y] = ysum[y];
				if(i+1<YADIF_MT_THREADS) {
					ysum[y]+=ystep[y];
					threads->args[i].y2[y] = ysum[y];
				}
				else { // last thread, different if (height % YADIF_MT_THREADS != 0)
					threads->args[i].y2[y] = buf_imgs[3]->height[y];
				}
				printf("start line: %i end line: %i plane: %i\n",threads->args[i].y[y],threads->args[i].y2[y],y);
			}
			
			pthread_create(&threads->threads[i], &attr, VaapiYadif_thread, &threads->args[i]);
		}
		printf("created threads\n");
#endif
	}
	else {
#if YADIF_TICKS
		tick1a = GetMsTicks();
#endif
		CopyVAImg2YadifImg(display, src, buf_imgs[2]);
#if YADIF_TICKS
		tick1b = GetMsTicks();
#endif
#if YADIF_MT
		//printf("wakeup waiting threads\n");
		pthread_cond_broadcast(&threads->cond);
#endif
	}
#if YADIF_TICKS
	tick2 = GetMsTicks();
#endif
#if YADIF_MT
	//printf("barrier wait main\n");
	pthread_barrier_wait(&threads->barrier);
#else
    for (i = 0; i < 3; i++) {
        int w = buf_imgs[3]->width[i];
        int h = buf_imgs[3]->height[i];
        int refs = buf_imgs[3]->pitch[i];

        for (y = 0; y < h; y++) {
            if ((y ^ tff) & 1) {
                uint8_t *prev = &buf_imgs[0]->yuv[i][y*refs];
                uint8_t *cur  = &buf_imgs[1]->yuv[i][y*refs];
                uint8_t *next = &buf_imgs[2]->yuv[i][y*refs];
#if HAVE_SSSE3
                vdr_plg_shddev_yadif_filter_line_ssse3(&buf_imgs[4]->yuv[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#elif HAVE_SSE2
                vdr_plg_shddev_yadif_filter_line_sse2(&buf_imgs[4]->yuv[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#else
                filter_line_c(&buf_imgs[4]->yuv[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#endif
                memcpy(&buf_imgs[3]->yuv[i][y*refs], cur, w);
            } else {
                uint8_t *prev = &buf_imgs[0]->yuv[i][y*refs];
                uint8_t *cur  = &buf_imgs[1]->yuv[i][y*refs];
                uint8_t *next = &buf_imgs[2]->yuv[i][y*refs];
#if HAVE_SSSE3
                vdr_plg_shddev_yadif_filter_line_ssse3(&buf_imgs[3]->yuv[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#elif HAVE_SSE2
                vdr_plg_shddev_yadif_filter_line_sse2(&buf_imgs[3]->yuv[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#else
                filter_line_c(&buf_imgs[3]->yuv[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
#endif
                memcpy(&buf_imgs[4]->yuv[i][y*refs], cur, w);
            }
        }
    }
#if HAVE_MMX
    __asm__ volatile("emms \n\t" : : : "memory");
#endif
#endif // YADIF_MT
#if YADIF_TICKS
	tick3 = GetMsTicks();
#endif
	CopyYadifImg2VAImg(display,buf_imgs[3],dst1);
	CopyYadifImg2VAImg(display,buf_imgs[4],dst2);
#if YADIF_TICKS
	tick4 = GetMsTicks();
#endif
	tmp = buf_imgs[0];
	buf_imgs[0] = buf_imgs[1];
	buf_imgs[1] = buf_imgs[2];
	buf_imgs[2] = tmp;
#if YADIF_TICKS
	printf("Yadif Times: %i %i %i - %i\n", tick2-tick1, tick3-tick2, tick4-tick3, tick4-tick1);
#endif
}

